#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <SDL/SDL.h>
#include <SDL/SDL_thread.h>
#include <SDL/SDL_ttf.h>

#include "irmp3io.h"

#define WITH_PREAMP_CODE 1

struct	icon
{
	int	x,y;
	char	*pID;
	char	*pFileName;
	char	*pAltFileName;
	SDL_Surface	*Image;
	SDL_Surface	*AltImage;
};

#define	LB_MODE_ALIGNMASK	3	//
#define	LB_MODE_LEFTALIGN	1	// These are mutually exclusive.
#define	LB_MODE_CENTERALIGN	2	//
#define	LB_MODE_RIGHTALIGN	3	//

#define	LB_MODE_ANTI	4

struct label
{
	int	x, y;
	char	*pID;
	TTF_Font	*font;
	SDL_Color	fg, bg;
	int	mode;

// the following are 'runtime' things...
	int	maxwidth;
	int	maxheight;
};

struct progressbar
{
	int	x, y, w, h;
	char	*pID;
	SDL_Color	fg, bg;
};

int RenderLabel(SDL_Surface *, char *pID, char *pText);
int BlankLabel(SDL_Surface *, char *pID);
int	RenderStaticLabels(SDL_Surface* surface);
int RenderStatusText(SDL_Surface* surface, struct label *pLabel, char *pText);
int PlaceIcon(char *name, SDL_Surface *surface, int Alt);
int	ReadConfig(char *, SDL_Surface *);
int	keyloop(void *data);
int SendCommand(int iSocket, char *pCommand);
int	RenderProgressBar(SDL_Surface *surface, struct progressbar *pPB, int value);

#define MAX_KEYS	512	// maximum number of a single keysym.

char	bQuit = 0;
SDL_Surface *screen;

// config items
char	*keycommands[MAX_KEYS];
char	*server;
char	*defaultfont;	//name of default
int		defaultfontsize;
struct icon	icons[100];
struct label labels[100];
struct progressbar progressbars[100];
int	iIconCount;
int	iLabelCount;
int	iProgressBarCount;

int main (int argc, char *argv[])
{
	int iSocket;
	SDL_Surface *statussurface;
	fd_set	fds;
	struct timeval	tv;
	int	i;

	/* Initialize the SDL library */
	if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
		fprintf(stderr, "Unable to init SDL: %s\n", SDL_GetError());
		exit(1);
	}

	/* Initialize the TTF library */
	if ( TTF_Init() < 0 ) {
			fprintf(stderr, "Couldn't initialize TTF: %s\n",SDL_GetError());
			exit(2);
	}
	atexit(SDL_Quit);

	screen = SDL_SetVideoMode(640, 480, 16, SDL_ANYFORMAT);
	if ( screen == NULL ) {
		fprintf(stderr, "Couldn't set 640x480x16 video mode: %s\n",
		SDL_GetError());
		exit(1);
	}

	//Blank the screen.
	SDL_FillRect(screen, NULL, SDL_MapRGBA(screen->format, 0, 0, 0, 0));

	statussurface = SDL_CreateRGBSurface(SDL_HWSURFACE, 640, 480, 16, 0, 0, 0, 0);

	// defaults for the configuration...
	server = "localhost";
	defaultfont = "/usr/share/fonts/default/TrueType/comic.ttf";	// works on RH 7.2
	defaultfontsize = 20;
	iIconCount = 0;
	for (i=0; i< MAX_KEYS; i++)
		keycommands[i] = 0;

	if (!ReadConfig("irmp3_sdl_interface.conf", statussurface))
		exit(1);

	RenderStaticLabels(statussurface);
	SDL_BlitSurface(statussurface, NULL, screen, NULL);
	SDL_UpdateRect(screen, 0, 0, 0, 0);

	iSocket = irmp3_io_connect(server);
	if (iSocket > 0)
	{
		SendCommand(iSocket, "query volume");
		SendCommand(iSocket, "query title");
		SendCommand(iSocket, "query artist");
		SendCommand(iSocket, "query album");
		SendCommand(iSocket, "query year");
		SendCommand(iSocket, "query genre");
		SendCommand(iSocket, "query comment");
		SendCommand(iSocket, "query mp3file");
		SendCommand(iSocket, "query status");
#ifdef WITH_PREAMP_CODE
		SendCommand(iSocket, "mixer channel CD +0");
		SendCommand(iSocket, "mixer channel line +0");
		SendCommand(iSocket, "mixer channel line1 +0");
		SendCommand(iSocket, "mixer channel pcm +0");
#endif
		SDL_CreateThread(keyloop, &iSocket);
		while (bQuit == 0)
		{
			FD_ZERO(&fds);
			FD_SET(iSocket, &fds);
	
			tv.tv_sec = 0;
			tv.tv_usec = 500000; // 500mS
			if (select(iSocket + 1, &fds, 0, 0, &tv) == 0)
			{
			}
			if (FD_ISSET(iSocket, &fds))
			{
				irmp3_io_data(iSocket, statussurface);
//				SDL_BlitSurface(statussurface, NULL,  screen, NULL);
//				SDL_Flip(screen);
			}
		}
		close (iSocket);
	}
	SDL_Quit();
	return 1;
}

// fix this...
int SendCommand(int iSocket, char *pCommand)
{
	int iRetstat;
	if ((iRetstat = send(iSocket, pCommand, strlen(pCommand), 0)) >= 0)
		iRetstat = send(iSocket, "\r\n", 2, 0);
	return iRetstat;
}

int	keyloop(void *data)
{
	int	iSocket = *((int *)data);
	SDL_Event	event;
	while (bQuit == 0)
	{
		SDL_WaitEvent(&event);
		switch(event.type)
		{
			case SDL_QUIT:
				bQuit = 1;
				break;
			case SDL_KEYDOWN:
				if (event.key.keysym.sym < MAX_KEYS)
				{
					if (keycommands[event.key.keysym.sym])
					{
						SendCommand(iSocket, keycommands[event.key.keysym.sym]);
					}
				}
				break;
			default:
				break;
		}
	}
	return 1;
}

void irmp3_io_callback_artist(void *data, char *text)
{
	SDL_Surface *surface = (SDL_Surface *)data;
	RenderLabel(surface, "@ARTIST", text);
}

void irmp3_io_callback_album(void *data, char *text)
{
	SDL_Surface *surface = (SDL_Surface *)data;
	RenderLabel(surface, "@ALBUM", text);
}

void irmp3_io_callback_title(void *data, char *text)
{
	SDL_Surface *surface = (SDL_Surface *)data;
	RenderLabel(surface, "@TITLE", text);
}

void irmp3_io_callback_year(void *data, char *text)
{
	SDL_Surface *surface = (SDL_Surface *)data;
	RenderLabel(surface, "@YEAR", text);
}

void irmp3_io_callback_comment(void *data, char *text)
{
	SDL_Surface *surface = (SDL_Surface *)data;
	RenderLabel(surface, "@COMMENT", text);
}

void irmp3_io_callback_endofsong(void *data)
{
	SDL_Surface *surface = (SDL_Surface *)data;
	BlankLabel(surface, "@FILE");
}

void irmp3_io_callback_file(void *data, char *text)
{
	SDL_Surface *surface = (SDL_Surface *)data;
	BlankLabel(surface, "@FILE");
	RenderLabel(surface, "@FILE", text);
}

void irmp3_io_callback_playlist(void *data, char *text)
{
	SDL_Surface *surface = (SDL_Surface *)data;
	RenderLabel(surface, "@PLAYLIST", text);
}

void irmp3_io_callback_volume(void *data, int iVolume)
{
	SDL_Surface *surface = (SDL_Surface *)data;
	char	caText[20];
	int	i;
	for (i=0; i < iProgressBarCount; i++)
	{
		if (strcmp(progressbars[i].pID, "@VOLUME") == 0)
		{
			RenderProgressBar(surface, &progressbars[i], iVolume);
			break;
		}
	}

	sprintf(caText, "%d %%", iVolume);
	RenderLabel(surface, "@VOLUME", caText);

}

void irmp3_io_callback_time(void *data, int amount_done, int amount_todo)
{
	SDL_Surface *surface = (SDL_Surface *)data;
	char	caText[20];
	int	i;

	if (amount_todo == 0 && amount_done == 0)	// prevent divide by zero.
		return;
	for (i=0; i < iProgressBarCount; i++)
	{
		if (strcmp(progressbars[i].pID, "@TIME") == 0)
		{
			RenderProgressBar(surface, &progressbars[i], (amount_done * 100) / (amount_done + amount_todo));
			break;
		}
	}

	sprintf(caText, "%d:%02d (%d:%02d)", amount_done / 60, amount_done % 60, amount_todo / 60, amount_todo % 60);
	RenderLabel(surface, "@TIME", caText);
}

void irmp3_io_callback_genre(void *data, char *text)
{
	SDL_Surface *surface = (SDL_Surface *)data;
	RenderLabel(surface, "@GENRE", text);
}

void irmp3_io_callback_beginofsong(void *data)
{
	SDL_Surface *surface = (SDL_Surface *)data;

	PlaceIcon("play", surface, 0);
	BlankLabel(surface, "@ARTIST");
	BlankLabel(surface, "@ALBUM");
	BlankLabel(surface, "@TITLE");
	BlankLabel(surface, "@GENRE");
}

void irmp3_io_callback_pause(void *data)
{
	SDL_Surface *surface = (SDL_Surface *)data;
	PlaceIcon("pause", surface, 0);
}

void irmp3_io_callback_stop(void *data)
{
	SDL_Surface *surface = (SDL_Surface *)data;
	PlaceIcon("stop", surface, 0);
}

void irmp3_io_callback_channel(void *data, char *text, int level)
{
#ifdef WITH_PREAMP_CODE
	SDL_Surface *surface = (SDL_Surface *)data;
	if (level > 0)
		PlaceIcon(text, surface, 0);
	else
		PlaceIcon(text, surface, 1);
#endif
	return;

}

void irmp3_io_callback_mute(void *data, int status)
{
}


int PlaceIcon(char *name, SDL_Surface *surface, int Alt)
{
	int	i;
	SDL_Rect	dest;
	for (i = 0; i < iIconCount; i++)
	{
		if (strcmp(name, icons[i].pID) == 0)
		{
			dest.x = icons[i].x;
			dest.y = icons[i].y;
			if (Alt)
				SDL_BlitSurface(icons[i].AltImage, NULL, surface, &dest);
			else
				SDL_BlitSurface(icons[i].Image, NULL, surface, &dest);

			SDL_BlitSurface(surface, &dest, screen, &dest);
			SDL_UpdateRects(screen, 1, &dest);
			return 1;
			break;
		}
	}
	return 0;
}

int	RenderProgressBar(SDL_Surface *surface, struct progressbar *pPB, int value)
{
	SDL_Rect	pb_dark;
	SDL_Rect	pb_light;

	pb_dark.x = pPB->x;
	pb_dark.y = pPB->y;
	pb_dark.w = (value * pPB->w) / 100;
	pb_dark.h = pPB->h;

	SDL_FillRect(surface, &pb_dark, SDL_MapRGBA(surface->format, pPB->fg.r, pPB->fg.g, pPB->fg.b, 0));

	pb_light.x = pb_dark.w + pPB->x;
	pb_light.y = pPB->y;
	pb_light.w = pPB->w - pb_dark.w;
	pb_light.h = pPB->h;

	SDL_FillRect(surface, &pb_light, SDL_MapRGBA(surface->format, pPB->bg.r, pPB->bg.g, pPB->bg.b, 0));

	pb_dark.w = pPB->w;
	SDL_BlitSurface(surface, &pb_dark, screen, &pb_dark);
	SDL_UpdateRects(screen, 1, &pb_dark);
	return 1;
}

int	BlankLabel(SDL_Surface* surface, char *pID)
{
	int i;
	SDL_Rect	back;
	for (i = 0; i < iLabelCount; i++)
	{
		if (strcmp(labels[i].pID, pID) == 0)
		{
			switch(labels[i].mode & LB_MODE_ALIGNMASK)
			{
				case LB_MODE_LEFTALIGN:
					back.x = labels[i].x;
					break;
				case LB_MODE_RIGHTALIGN:
					back.x = labels[i].x - labels[i].maxwidth;
					break;
				case LB_MODE_CENTERALIGN:
					back.x = labels[i].x - (labels[i].maxwidth/2);
					break;
			}
			back.y = labels[i].y;
			back.w = labels[i].maxwidth;
			back.h = labels[i].maxheight;
			SDL_FillRect(surface, &back, SDL_MapRGBA(surface->format, 0, 0, 0, 0));
			SDL_BlitSurface(surface, &back, screen, &back);
			SDL_UpdateRects(screen, 1, &back);
			return i;
		}
	}
	return 0;
}

int	RenderStaticLabels(SDL_Surface* surface)
{
	int i;
	for (i = 0; i < iLabelCount; i++)
	{
		if (*labels[i].pID == '@')
			continue;
		RenderStatusText(surface, &(labels[i]), labels[i].pID);
	}
	return 0;
}

int	RenderLabel(SDL_Surface* surface, char *pID, char *pText)
{
	int i;
	for (i = 0; i < iLabelCount; i++)
	{
		if (strcmp(labels[i].pID, pID) == 0)
		{
			RenderStatusText(surface, &(labels[i]), pText);
			return i;
		}
	}
	return 0;
}

int RenderStatusText(SDL_Surface* surface, struct label *pLabel, char *pText)
{
	SDL_Rect back;
	SDL_Surface* text;

	if (! pLabel->mode & LB_MODE_ANTI)
		text = TTF_RenderText_Shaded(pLabel->font, pText, pLabel->fg, pLabel->bg);
	else
		text = TTF_RenderText_Blended(pLabel->font, pText, pLabel->fg);
	
	if (text != NULL)
	{
		if ((pLabel->mode & LB_MODE_ALIGNMASK) == LB_MODE_CENTERALIGN)
		{
			back.x = pLabel->x - (text->w/2);
		}
		else if ((pLabel->mode & LB_MODE_ALIGNMASK) == LB_MODE_RIGHTALIGN)
		{
			back.x = pLabel->x - text->w;
		}
		else
		{
			back.x = pLabel->x;
		}

		back.y = pLabel->y;
		back.h = text->h;
		if (back.h > pLabel->maxheight)
			pLabel->maxheight = back.h;

		back.w = text->w;
		if (back.w > pLabel->maxwidth)
			pLabel->maxwidth = back.w;

		if ((pLabel->mode & LB_MODE_ALIGNMASK) == LB_MODE_CENTERALIGN)
		{
			back.x = pLabel->x - (text->w / 2);
		}
		else if ((pLabel->mode & LB_MODE_ALIGNMASK) == LB_MODE_RIGHTALIGN)
		{
			back.x = pLabel->x - text->w;
		}
		else
		{
			back.x = pLabel->x;
		}

		SDL_BlitSurface(text, NULL, surface, &back);

		back.w = pLabel->maxwidth;
		SDL_BlitSurface(surface, &back, screen, &back);
		SDL_UpdateRects(screen, 1, &back);

		SDL_FreeSurface(text);
	}
	return 1;
}

int	ReadConfig(char *file, SDL_Surface *surface)
{
	char	caLine[512];
	char	*pArgs[20];
	int	iArgCount;
	int iLineCount = 0;
	FILE	*fp = fopen(file, "r");
	if (!fp)
	{
		perror(file);
		return 0;
	}
	while (fgets(caLine, sizeof(caLine), fp))
	{
		iLineCount++;
		ParseLine(caLine, pArgs, &iArgCount, 20);
		if (pArgs[0][0] == '#')
			continue;
		if (iArgCount > 0)
		{
			if (strcmp(pArgs[0], "key") == 0)
			{
				// FIXME:
				// for now you have to put the actual keycode number from /usr/include/SDL/SDK_keysym.h
				if (iArgCount >= 3)
				{
					int iKey = atoi(pArgs[1]);
					if (iKey > 0 && iKey < MAX_KEYS)
					{
						keycommands[iKey] = strdup(pArgs[2]);
//						printf("Key: %d, value '%s'\n", iKey, keycommands[iKey]);
					}
					else
						fprintf(stderr, "line %d: Keycode out of range\n", iLineCount);
				}
				else
					fprintf(stderr, "line %d: invalid Syntax\n", iLineCount);
			}
			else if (strcmp(pArgs[0], "server") == 0)
			{
				if (iArgCount >= 2)
					server = strdup(pArgs[1]);
				else
					fprintf(stderr, "line %d: invalid Syntax\n", iLineCount);
			}
			else if (strcmp(pArgs[0], "defaultfont") == 0)
			{
				if (iArgCount >= 3)
					defaultfontsize = atoi(pArgs[2]);
				if (iArgCount >= 2)
					defaultfont = strdup(pArgs[1]);
				else
					fprintf(stderr, "line %d: invalid Syntax\n", iLineCount);
			}
			else if (strcmp(pArgs[0], "label") == 0)
			{
				// syntax:
				// label x y "value" [align] [font] [size] [color r] [color b] [color g]
				if (iArgCount >= 4) // minimum number of args
				{
					labels[iLabelCount].x = atoi(pArgs[1]);
					labels[iLabelCount].y = atoi(pArgs[2]);
					labels[iLabelCount].maxwidth = 0;
					labels[iLabelCount].maxheight = 0;
					labels[iLabelCount].mode = LB_MODE_ANTI;

					if (iArgCount >= 5)	// align
					{
						if (strcmp(pArgs[4], "left") == 0)	labels[iLabelCount].mode |= LB_MODE_LEFTALIGN;
						else if (strcmp(pArgs[4], "right") == 0)	labels[iLabelCount].mode |= LB_MODE_RIGHTALIGN;
						else if (strcmp(pArgs[4], "center") == 0)	labels[iLabelCount].mode |= LB_MODE_CENTERALIGN;
						else
							fprintf(stderr, "line %d: invalid alignment\n", iLineCount);
					}
					if (iArgCount >= 6)	// font
					{
						if (strcmp(pArgs[5], "default") == 0)
							pArgs[5] = defaultfont;

						if (iArgCount >= 7)	// size
						{
							labels[iLabelCount].font = TTF_OpenFont(pArgs[5], atoi(pArgs[6]));
						}
						else
							labels[iLabelCount].font = TTF_OpenFont(pArgs[5], defaultfontsize);
					}
					else
						labels[iLabelCount].font = TTF_OpenFont(defaultfont, defaultfontsize);

					labels[iLabelCount].fg.r = 255;
					labels[iLabelCount].fg.g = 255;
					labels[iLabelCount].fg.b = 255;
					labels[iLabelCount].bg.r = 0;
					labels[iLabelCount].bg.g = 0;
					labels[iLabelCount].bg.b = 0;
					if (iArgCount >= 8) labels[iLabelCount].fg.r = atoi(pArgs[7]); // Red
					if (iArgCount >= 9) labels[iLabelCount].fg.g = atoi(pArgs[8]); // Green
					if (iArgCount >= 10) labels[iLabelCount].fg.b = atoi(pArgs[9]);	// Blue

					if (!labels[iLabelCount].font)
						fprintf(stderr, "line %d: Font Error\n", iLineCount);
					else
					{
						labels[iLabelCount].pID = strdup(pArgs[3]);
						iLabelCount++;
					}
				}
				else
					fprintf(stderr, "line %d: invalid Syntax\n", iLineCount);
			}
			else if (strcmp(pArgs[0], "icon") == 0)
			{
				// syntax:
				// icon x y id name [altname]
				if (iArgCount >= 5)
				{
					icons[iIconCount].x = atoi(pArgs[1]);
					icons[iIconCount].y = atoi(pArgs[2]);
					icons[iIconCount].pID = strdup(pArgs[3]);
					icons[iIconCount].pFileName = strdup(pArgs[4]);
					icons[iIconCount].Image = SDL_LoadBMP(icons[iIconCount].pFileName);
					if (!icons[iIconCount].Image)
						fprintf(stderr, "line %d: unable to load icon %s\n", iLineCount, icons[iIconCount].pFileName);
					if (iArgCount >= 6)
					{
						icons[iIconCount].pAltFileName = strdup(pArgs[5]);
						icons[iIconCount].AltImage = SDL_LoadBMP(icons[iIconCount].pAltFileName);
						if (!icons[iIconCount].AltImage)
							fprintf(stderr, "line %d: unable to load icon %s\n", iLineCount, icons[iIconCount].pAltFileName);
					}
					else
						icons[iIconCount].pAltFileName = 0;
					iIconCount++;
				}
				else
					fprintf(stderr, "line %d: invalid Syntax\n", iLineCount);
			}
			else if (strcmp(pArgs[0], "progressbar") == 0)
			{
				if (iArgCount >= 6)
				{
					progressbars[iProgressBarCount].x = atoi(pArgs[1]);
					progressbars[iProgressBarCount].y = atoi(pArgs[2]);
					progressbars[iProgressBarCount].w = atoi(pArgs[3]);
					progressbars[iProgressBarCount].h = atoi(pArgs[4]);
					progressbars[iProgressBarCount].pID = strdup(pArgs[5]);

					progressbars[iProgressBarCount].fg.r = 0;
					progressbars[iProgressBarCount].fg.g = 150;
					progressbars[iProgressBarCount].fg.b = 150;
					progressbars[iProgressBarCount].bg.r = 0;
					progressbars[iProgressBarCount].bg.g = 50;
					progressbars[iProgressBarCount].bg.b = 50;

					if (iArgCount >=7) progressbars[iProgressBarCount].fg.r = atoi(pArgs[6]); // Red
					if (iArgCount >=8) progressbars[iProgressBarCount].fg.g = atoi(pArgs[7]); // Green
					if (iArgCount >=9) progressbars[iProgressBarCount].fg.b = atoi(pArgs[8]); // Blue
					if (iArgCount >=10) progressbars[iProgressBarCount].bg.r = atoi(pArgs[9]); // Red
					if (iArgCount >=11) progressbars[iProgressBarCount].bg.g = atoi(pArgs[10]); // Green
					if (iArgCount >=12) progressbars[iProgressBarCount].bg.b = atoi(pArgs[11]); // Blue
					iProgressBarCount++;
				}
				else
					fprintf(stderr, "line %d: invalid Syntax\n", iLineCount);
			}
		}
	}
	fclose(fp);
	return 1;
}
